#include <gtk/gtk.h>
#include <antiprism/antiprism.h>
#include "named_model_list.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_SIGACTION
#include <signal.h>
#endif

#ifndef RESDIR
#define RESDIR .
#endif

enum {
   MODEL_NAME=0,
   MODEL_DESC,
   NUM_MODEL_COLS
};

enum {
   LOAD_MODEL_NAME=0,
   LOAD_MODEL_DESC,
   NUM_LOAD_MODEL_COLS
};

GtkListStore *load_model_store;

vector<string> get_res_file_paths(const char *fname)
{
   vector<string> dirs;
   dirs.push_back(fname);
   dirs.push_back(string(RESDIR)+"/"+fname);
   return dirs;
}


void init_gv()
{
   printf("(merge-ap World { concave })\n");
}


int write_gv_geom(const char *name, const char *suffix, char *errmsg=0)
{
   col_geom_v geom;
   int ret = geom.read(name, errmsg);
   if(ret) {
      printf("(geometry %s%s%s {\n", name, *suffix?"_":"", suffix);
      geom.write();
      printf("} )\n");
      fflush(stdout);
   }
   return ret;
}

void delete_gv_geom(const char *name)
{
   printf("(delete %s)\n", name);
   fflush(stdout);
}

void clear_models()
{
   GtkTreeIter iter;
   gboolean valid;
   valid = gtk_tree_model_get_iter_first(
         GTK_TREE_MODEL(load_model_store), &iter);
   while (valid) {
      gchar *name;
      gtk_tree_model_get(GTK_TREE_MODEL(load_model_store), &iter,
                          LOAD_MODEL_NAME, &name, -1);
      delete_gv_geom(name);
      g_free(name);
      valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(load_model_store), &iter);
   }

   gtk_list_store_clear(GTK_LIST_STORE(load_model_store));
}

bool name_in_list(const char *name)
{
   bool in_list = false;
   GtkTreeIter iter;
   gboolean valid = gtk_tree_model_get_iter_first(
         GTK_TREE_MODEL(load_model_store), &iter);
   while(valid) {
      gchar *list_name;
      gtk_tree_model_get(GTK_TREE_MODEL(load_model_store), &iter,
                          LOAD_MODEL_NAME, &list_name, -1);
      if(strcmp(name, list_name)==0)
         in_list = true;
      g_free(list_name);
      if(in_list)
         break;

      valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(load_model_store), &iter);
   }
   return in_list;
}


void update_load_list_from_text_entry(GtkEntry *entry)
{
   const gchar *name = gtk_entry_get_text(GTK_ENTRY(entry));
   if(*name && !name_in_list(name)) {
      GtkTreeIter iter;
      gtk_list_store_append(load_model_store, &iter);
      gtk_list_store_set(load_model_store, &iter, LOAD_MODEL_NAME, name, -1);
      gtk_entry_set_text(GTK_ENTRY(entry), "");
   }
}


#ifdef HAVE_SIGACTION
extern "C" void clear_models_on_signal(int sig)
{
   clear_models();

   struct sigaction new_action, old_action;
   new_action.sa_handler = SIG_DFL;
   sigemptyset (&new_action.sa_mask);
   new_action.sa_flags = 0;

   if(old_action.sa_handler != SIG_IGN)
      sigaction (sig, &new_action, NULL);
   raise(sig);
}

void set_signal_handlers()
{
   struct sigaction new_action, old_action;
   new_action.sa_handler = clear_models_on_signal;
   sigemptyset (&new_action.sa_mask);
   new_action.sa_flags = 0;

   if (old_action.sa_handler != SIG_IGN)
      sigaction (SIGHUP, &new_action, NULL);
   sigaction (SIGTERM, NULL, &old_action);
}
#endif


extern "C" G_MODULE_EXPORT gboolean on_MAIN_WINDOW_destroy_event(
      GtkObject * /*object*/, gpointer /*data*/)
{
   clear_models();
   gtk_main_quit();
   return TRUE;
}

extern "C" G_MODULE_EXPORT gboolean on_MAIN_WINDOW_delete_event(
      GtkObject * /*object*/, gpointer /*data*/)
{
   clear_models();
   gtk_main_quit();
   return TRUE;
}

extern "C" G_MODULE_EXPORT void on_BUTTON_HELP_OK_clicked(gpointer data,
      GtkButton * /*button*/)
{
   gtk_widget_hide(GTK_WIDGET(data));
}

extern "C" G_MODULE_EXPORT void on_BUTTON_HELP_clicked(gpointer data,
      GtkButton * /*button*/)
{
   gtk_widget_show_all(GTK_WIDGET(data));
}

extern "C" G_MODULE_EXPORT void on_MESSAGE_ABOUT_response( GtkDialog *dialog,
      guint /*id*/)
{
   gtk_widget_hide(GTK_WIDGET(dialog));
}

extern "C" G_MODULE_EXPORT void on_BUTTON_ABOUT_clicked(gpointer data,
      GtkButton * /*button*/)
{
   gtk_widget_show_all(GTK_WIDGET(data));
}

extern "C" G_MODULE_EXPORT void on_ADD_MODEL_TEXT_activate(GtkEntry *entry,
      gpointer /*data*/)
{
   update_load_list_from_text_entry(entry);
}

extern "C" G_MODULE_EXPORT void on_BUTTON_ADD_clicked(gpointer data,
      GtkButton * /*button*/)
{
   update_load_list_from_text_entry(GTK_ENTRY(data));
}

extern "C" G_MODULE_EXPORT void on_BUTTON_APPLY_clicked(gpointer data,
      GtkButton * /*button*/)
{
   rand_gen rnd;
   rnd.time_seed();
   string batch = itostr(rnd(100000));   // convenient "unique" id

   GtkTreeIter iter;
   gboolean valid;
   valid = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(data), &iter);
   while (valid) {
      gchar *name;
      gtk_tree_model_get (GTK_TREE_MODEL(data), &iter,
                          LOAD_MODEL_NAME, &name, -1);
      delete_gv_geom(name);
      write_gv_geom(name, batch.c_str());
      g_free(name);
      valid = gtk_tree_model_iter_next (GTK_TREE_MODEL(data), &iter);
   }

   gtk_list_store_clear(GTK_LIST_STORE(data));
}


extern "C" G_MODULE_EXPORT void on_BUTTON_CLEAR_clicked(
      GtkButton * /*button*/, gpointer /*data*/)
{
   clear_models();
}


extern "C" G_MODULE_EXPORT void on_BUTTON_CLOSE_clicked(
      GtkButton * /*button*/, gpointer /*data*/)
{
   clear_models();
   gtk_main_quit();
}


extern "C" G_MODULE_EXPORT void on_MODEL_TREE_row_activated(gpointer data,
      GtkTreePath *path, GtkTreeViewColumn * /* col*/, GtkTreeView *treeview )
{
    GtkTreeModel *model = gtk_tree_view_get_model(treeview);
    GtkTreeIter iter, child;
    if (gtk_tree_model_get_iter(model, &iter, path) &&
        !gtk_tree_model_iter_children(model, &child, &iter) ) {
       gchar *name;
       gchar *desc;
       gtk_tree_model_get(model, &iter,
             MODEL_NAME, &name, MODEL_DESC, &desc, -1);
       if(!name_in_list(name)) {
          GtkListStore *liststore = GTK_LIST_STORE(G_OBJECT(data));
          GtkTreeIter list_iter;
          gtk_list_store_append(liststore, &list_iter);
          gtk_list_store_set(liststore, &list_iter,
                       LOAD_MODEL_NAME, name,
                       LOAD_MODEL_DESC, desc, -1);
       }

       g_free(name);
       g_free(desc);
    }
}

extern "C" G_MODULE_EXPORT void on_LOAD_MODEL_LIST_row_activated(
      GtkTreeView *treeview, GtkTreePath *path, GtkTreeViewColumn * /* col*/,
      gpointer * /*data*/ )
{
    GtkTreeModel *model = gtk_tree_view_get_model(treeview);
    GtkTreeIter iter;
    if (gtk_tree_model_get_iter(model, &iter, path)) {
       gchar *name;
       gtk_tree_model_get(model, &iter, MODEL_NAME, &name, -1);
       delete_gv_geom(name);
       gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
       g_free(name);
    }
}

extern "C" G_MODULE_EXPORT void on_LOAD_MODEL_STORE_row_changed(
      GtkTreeModel *treemodel, GtkTreePath * /*path*/, GtkTreeIter *iter)
{
    gchar *name;
    gtk_tree_model_get(treemodel, iter, MODEL_NAME, &name, -1);
    char errmsg[MSG_SZ];
    if(!write_gv_geom(name, "", errmsg)) {
       GtkMessageDialog *dialog = GTK_MESSAGE_DIALOG(gtk_message_dialog_new(
             NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
             GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
             "Model name error '%s':\n%s", name, errmsg) );
       gtk_dialog_run(GTK_DIALOG(dialog));
       gtk_widget_destroy(GTK_WIDGET(dialog));
       gtk_list_store_remove(GTK_LIST_STORE(treemodel), iter);
    }
    g_free(name);
}

void set_up_models_add(GtkTreeStore *treestore, GtkTreeIter *iter,
      const char *model_list[][2])
{
   GtkTreeIter child, child2;
   gtk_tree_store_append(treestore, &child, iter);
   gtk_tree_store_set (treestore, &child,
                       MODEL_NAME, model_list[0][0],
                       MODEL_DESC, "", -1);
   int i=0;
   while(*model_list[++i][0]) {   // quit on null string at end of list
      gtk_tree_store_append(treestore, &child2, &child);
      gtk_tree_store_set (treestore, &child2,
                          MODEL_NAME, model_list[i][0],
                          MODEL_DESC, model_list[i][1], -1);
   }
}

void set_up_models(GtkTreeStore *treestore)
{
   GtkTreeIter iter;
   gtk_tree_model_get_iter_first(GTK_TREE_MODEL(treestore), &iter);
   gtk_tree_store_append(treestore, &iter, NULL);
   gtk_tree_store_set (treestore, &iter,
                       MODEL_NAME, "Polyhedron Models",
                       MODEL_DESC, "", -1);
   set_up_models_add(treestore, &iter, model_names_common);
   set_up_models_add(treestore, &iter, model_names_uniform);
   set_up_models_add(treestore, &iter, model_names_uniform_dual);
   set_up_models_add(treestore, &iter, model_names_uniform_compound);
   set_up_models_add(treestore, &iter, model_names_johnson);
   set_up_models_add(treestore, &iter, model_names_polygon);
   set_up_models_add(treestore, &iter, model_names_geodesic);
   set_up_models_add(treestore, &iter, model_names_geodesic_dual);
   set_up_models_add(treestore, &iter, model_names_standard);
   set_up_models_add(treestore, &iter, model_names_symmetry_ex);

}

int main (int argc, char *argv[])
{
        gtk_init (&argc, &argv);
        
        GtkBuilder *builder = gtk_builder_new();
        vector<string> paths = get_res_file_paths("antiprism_models.glade");
        vector<string>::const_iterator vi = paths.begin();
        while(vi!=paths.end() && !gtk_builder_add_from_file(
                                                 builder, vi->c_str(), NULL))
           ++vi;

        if(vi==paths.end()) {
           fprintf(stderr, "error: could not open UI definition file: %s",
                 paths[0].c_str());
           exit(1);
        }

        GtkWidget *window =
           GTK_WIDGET(gtk_builder_get_object(builder, "MAIN_WINDOW"));
        GtkTreeStore *treestore =
           GTK_TREE_STORE(gtk_builder_get_object(builder, "MODEL_TREE_STORE"));
        load_model_store =
           GTK_LIST_STORE(gtk_builder_get_object(builder, "LOAD_MODEL_STORE"));
        set_up_models(treestore);
        
        GtkTreeView *treeview =
           GTK_TREE_VIEW(gtk_builder_get_object(builder, "MODEL_TREE"));

        GError *err = 0; // to supress warnings
        paths = get_res_file_paths("icon_snu4_64.png");
        vi = paths.begin();
        while(vi!=paths.end() &&
              !gtk_window_set_default_icon_from_file(vi->c_str(), &err) ) {
           if(err != 0) {
              g_error_free (err);
              err = 0;
           }
           ++vi;
        }
        if(vi==paths.end())
           fprintf(stderr, "warning: could not open default icon file: %s\n",
                 paths[0].c_str());

        
        gtk_builder_connect_signals(builder, NULL);          
        g_object_unref(G_OBJECT(builder));
        
        GtkTreePath *root_path = gtk_tree_path_new_first();
        gtk_tree_view_expand_row(treeview, root_path, false);
        gtk_tree_path_free(root_path);

#ifdef HAVE_SIGACTION
        set_signal_handlers();
#endif

        gtk_widget_show(window);
        init_gv();
        gtk_main();
        
        return 0;
}

